/*
 *  M.Meadele mm@bzero.net Thu May 23 02:30:07 GMT 2002 
 * 
 * Thu May 23 02:30:17 GMT 2002
 *---
 * - first version
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "misc.h"

#include "cookies.h"

static void    cook_add_son(struct cook_tree *, struct cook_tree *);

static struct cook_tree        *cook_tree_create(char *);
static struct cook_tree_list   *cook_tree_list_alloc(void);
static struct cook_tree_list_e *cook_tree_list_e_alloc(void);
      
static struct cookie *cookie_alloc(void);
static void   cookie_destroy(struct cookie *);
      
static struct cookie_list *cookie_list_alloc(void);
static void                cookie_list_destroy(struct cookie_list *);
      

static int cookie_add(char *, char *, struct cookie_list *);
static int cookie_del(char *, struct cookie_list *);
      
static void cookie_get(char *, char *, int, struct cook_tree *);
static int cookie_len(char *, struct cook_tree *); 

struct cook_tree *cookie_tree;

/*                      *
 ** COOKIES NAME/VALUE **
 *                      */

static int 
   cookie_add(char *name, char *value, struct cookie_list *cl)
{
 int i; 
 struct cookie *fc;

 /*
  * no cookie, insert new one.
  */
 if( cl->sz == 0 ) {
    cl->head = cookie_alloc();
    cl->head->name = strdup(name);
    cl->head->value = strdup(value);
    cl->tail = cl->head;
    cl->sz++;
    return COOK_INSERT;
    }
 /*
  * existing cookie list, 
  * if 'name' is an existing cookie,
  * replace its value by 'value'.
  */
 fc = cl->head;
 for(i=0; i<cl->sz; i++, fc=fc->next) {
     if( strcmp(fc->name, name) == 0 ) {
        free(fc->value);
        fc->value = strdup(value);
        return COOK_REPLACE;
        }
     }
 /*
  * 'name' doesn't exist in cookie_list,
  * create new entrie.
  */
 cl->tail->next = cookie_alloc();
 cl->tail->next->name  = strdup(name);
 cl->tail->next->value = strdup(value);
 cl->tail->next->next  = NULL;
 cl->tail = cl->tail->next;
 cl->sz++;
 return COOK_INSERT;
}
 

static int
   cookie_del(char *name, struct cookie_list *cl)
{
 int i;
 struct cookie *bc, *c;

 bc = c = cl->head;

 /*
  * if 'name' is NULL,
  * erase any cookies for this ressource.
  */
 if( name == NULL ) {
    for(i=0; i<cl->sz; i++) {
        c = bc->next;
        cookie_destroy(bc);
        bc = c;
        }
    cl->sz = 0;
    return COOK_DELETE;
    }

 /*
  * else, erase the cookie named 'name'
  */
 for(i=0; i<cl->sz; i++, c=c->next) {
     if( strcmp(c->name, name) == 0 ) {
        if( c == cl->head )
           cl->head = cl->head->next;
        else {
           if( c == cl->tail )
              cl->tail = bc;     
           else 
              bc->next = c->next;
           }
        cookie_destroy(c);
        cl->sz--;
        return COOK_DELETE;
        }
     else
        bc = c;
     }

 /*
  * cookie not found.
  */
 return COOK_NOTFOUND;
}



/*                *
 ** MEMORY STUFF **
 *                */


static struct cookie *
    cookie_alloc()
{

 return (xmalloc(sizeof(struct cookie)));
}


static void
   cookie_destroy(struct cookie *c)
{
 
 free(c->name);
 free(c->value);
 free(c);
}
 

static struct cookie_list *
   cookie_list_alloc()
{

 return (xcalloc(1, sizeof(struct cookie_list)));
}


static void
   cookie_list_destroy(struct cookie_list *cl)
{
 int i;
 struct cookie *a, *b;

 a = cl->head;
 for(i=0; i<cl->sz; i++) {
     b = a->next;
     cookie_destroy(a);
     a = b;
     }
 free(cl);
}


static struct cook_tree *
    cook_tree_create(char *p)
{
 struct cook_tree *ct;

 ct = xcalloc(1, sizeof(struct cook_tree));
 ct->name = strdup(p);
 ct->cookie_list = cookie_list_alloc();
 ct->sons = cook_tree_list_alloc();

 return ct;
}


static struct cook_tree_list *
    cook_tree_list_alloc()
{

 return (xcalloc(1, sizeof(struct cook_tree_list)));
}


static struct cook_tree_list_e *
    cook_tree_list_e_alloc()
{

 return (xmalloc(sizeof(struct cook_tree_list_e)));
}


/*        *
 ** TREE **
 *        */


static void cookie_get(char *path, char *cvalue, int len, struct cook_tree *ct)
{
 int i;
 int wl = 0;
 char *s, *p, *elm;
 struct cook_tree_list_e *ctle;
 struct cookie *cake;

 ctle = ct->sons->head;
 s = path;

 if( ct->cookie_list->sz != 0 ) {
    cake = ct->cookie_list->head;
    for(i=0; i<ct->cookie_list->sz; i++, cake=cake->next) 
        wl += sprintf(cvalue+wl, "%s=%s; ", cake->name, cake->value);
    }

 if( *s == '\0' || (*s == '/' && *++s == '\0') )
    return;

 p = strchr(s, '/');
 if( p == NULL ) 
    p = s+strlen(s);

 if( ct->sons->sz > 0 ) {
    elm = wstrndup(s, (p-s));
    for(i=0; i<ct->sons->sz; i++, ctle=ctle->next) {
        if( strcmp(ctle->ln->name, elm) == 0 ) {
           cookie_get(p, cvalue+wl, len-wl, ctle->ln);
           break;
           }
         }
    free(elm);
    }
}


static int cookie_len(char *path, struct cook_tree *ct) 
{ 
 char vuf[2];
 int i;
 int wl = 0;
 char *s, *p, *elm;
 struct cookie *cake;
 struct cook_tree_list_e *ctle;

 ctle = ct->sons->head;
 s = path;

 if( ct->cookie_list->sz != 0 ) {
    cake = ct->cookie_list->head;
    for(i=0; i<ct->cookie_list->sz; i++, cake=cake->next) 
 /*
  * vuf is only used to avoid  memory leak in some snprintf() implementations.
  * replacinf vuf by NULL, results in huge memory leak !!!!!!
  */
        wl += snprintf(vuf, 0, "%s=%s; ", cake->name, cake->value); 
    }

 if( *s == '\0' || (*s == '/' && *++s == '\0') )
    return wl;

 p = strchr(s, '/');
 if( p == NULL ) 
    p = s+strlen(s);

 if( ct->sons->sz > 0 ) {
    elm = wstrndup(s, (p-s));
    for(i=0; i<ct->sons->sz; i++, ctle=ctle->next) {
        if( strcmp(ctle->ln->name, elm) == 0 ) {
           wl += cookie_len(p, ctle->ln);
           break;
           }
         }
    free(elm);
    }

 return wl;
}


static void cook_add_son(struct cook_tree *ct, struct cook_tree *son)
{
 
 if( ct->sons->head == NULL ) {
    ct->sons->head       = cook_tree_list_e_alloc();
    ct->sons->head->ln   = son;
    ct->sons->head->next = NULL;
    ct->sons->tail       = ct->sons->head;
    }
 else {
    ct->sons->tail->next       = cook_tree_list_e_alloc();
    ct->sons->tail->next->ln   = son;
    ct->sons->tail->next->next = NULL;
    ct->sons->tail             = ct->sons->tail->next;
    }
 ct->sons->sz++;
}


struct cook_tree *cook_tree_init()
{

 cookie_tree = cook_tree_create("/");
 return cookie_tree; 
}


void cook_tree_destroy(struct cook_tree *ct)
{
 int i;
 struct cook_tree_list_e *ctle, *nctle;

 ctle = ct->sons->head;
 for(i=0; i<ct->sons->sz; i++) {
     nctle = ctle->next;
     cook_tree_destroy(ctle->ln);
     free(ctle);
     ctle = nctle;
     }
 free(ct->name);
 cookie_list_destroy(ct->cookie_list);
 free(ct->sons);
 free(ct);
}


char *cookie_build(char *path, struct cook_tree *ct)
{
 int sz;
 char *cake;

 sz = cookie_len(path, ct);
 if( sz == 0 )
    return NULL;
 cake = xmalloc(sz+1);

 cookie_get(path, cake, sz, ct);
 cake[sz-2] = '\0';

 return cake;
}


/*
 *  cookie_set() with name == NULL means erase all cookies for this path
 *  cookie_set() with value == NULL means erase the cookie with value 'value',
 *               only for THIS ressource, not recursive
 */
void cookie_set(char *path, char *cname, char *cvalue, struct cook_tree *ct)
{
 int i, r;
 char *s, *p, *elm;
 struct cook_tree_list_e *ctle;
 
 s    = path;
 ctle = ct->sons->head;

 if( *s == '\0' || (*s == '/' && *++s == '\0') ) {
    if( cvalue == NULL )
       r = cookie_del(cname, ct->cookie_list);
    else
       r = cookie_add(cname, cvalue, ct->cookie_list);
    return;
    }
 p = strchr(s, '/');
 if( p == NULL ) {
    p = s;
    while( *p != '\0' )
          p++;
    }

 elm = wstrndup(s, (p-s));
 for(i=0; i<ct->sons->sz; i++, ctle=ctle->next) {
     if( strcmp(ctle->ln->name, elm) == 0 ) {
        cookie_set(p, cname, cvalue, ctle->ln);
        free(elm);
        return;
        }
     }
 cook_add_son(ct, cook_tree_create(elm));
 cookie_set(p, cname, cvalue, ct->sons->tail->ln);
 free(elm);
}


void cookie_print(char *p, struct cook_tree *ct)
{
 char *grocake;

 grocake = cookie_build(p, ct);
 printf("cookie:  %s = %s\n", p, (grocake ? grocake : "(no cookie for this ressource)"));
 free(grocake);
}


/*
 * parse and insert a cookie from 
 * a HTTP transaction
 */
int cookie_insert(char *petilu, struct cook_tree *ct)
{
 char *name, *value, *path;
 char *s, *p, *q, *e, *sep;
 
 path = NULL;
 p = s = strdup(petilu);
 SKIPSP(p);

 e = strchr(p, ';');
 if( e == NULL ) {
    e = p + strlen(p);
    path = "/";
    }
 else
    *e = '\0';

 sep = strchr(p, '=');
 if( sep == NULL ) {
    m_warn("invalid cookie");
    free(s);
    return -1;
    }
 *sep = '\0';
 name  = p;
 value = sep+1;

 if( path != NULL ) {
    cookie_set(path, name, value, ct);
    free(s);
    return 0;
    }

 p = ++e;
 for(;;) {
     SKIPSP(p);
     if( strncasecmp(p, "path", 4) == 0 ) {
        sep = strchr(p+4, '=');
        if( sep == NULL ) {
           m_warn("invalid cookie");
           free(s);
           return -1;
           }
        sep++;
        SKIPSP(sep);
        q = sep;
        if( *sep == '\"' ) {
           sep++;
           while( *q != '\"' && *q != '\0' )
                 q++;
           *q = '\0';
           }
        else {
           while( *q != ';' && *q != '\0' )
                 q++;
           *q = '\0';
           }
        path = sep;
        break;
        }
     e = strchr(p, ';');
     if( e == NULL ) {
        path = "/";
        break;
        }
     p = ++e;
     }

 cookie_set(path, name, value, ct);
 free(s);

 return 0;
}


/*
int main(int argc, char *argv[])
{
 struct cook_tree *ct;

 ct = cook_tree_init(); 

     cookie_insert("sessionhash=14005b2ecb534a1ef98c81fad9074f7e; path=/test",  ct);
 
 cookie_print("/",                ct);
 cookie_print("/test",            ct);

 cook_tree_destroy(ct);

 return 0;
}
*/
