/* M.Meadele mm@bzero.net Wed May  8 00:09:05 GMT 2002  
 *
 *---
 * Tue May  7 14:23:42 GMT 2002
 * - first version
 *---
 *
 *---
 * Fri May 24 03:44:50 GMT 2002
 * - added "Cookie:" header in http_build_default_hdr_list()
 *
 *---
 * Mon May 27 03:14:25 GMT 2002
 * - added http_cookie_insert()
 *
 *---
 * Mon May 27 18:42:52 GMT 2002
 * - added http_get_hdr_value()
 *
 * Wed May 29 00:00:01 GMT 2002
 *---
 * - removed http_set_method()
 * - removed http_get_method()
 */
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <netdb.h>
#include <netinet/in.h>

#include <errno.h>
#include <err.h>
#include <sysexits.h>

#include "misc.h"
#include "url.h"
#include "cookies.h"

#include "http.h"


#define CONNECT_TIMEOUT    8
#define CHUNK_SIZE         8192

static void h_sigalarm(int no);


#define HOSTINFO_BUILD_PORT(a, b)     ((a) = (b))  
static int   hostinfo_resolve_hostname(char *, struct host_info *);
static FILE *hostinfo_connect(struct host_info *);
static void  hostinfo_close(struct host_info *);

static void      http_free_chunks(msg_chunk *);

static size_t http_write_msg(struct http_msg *, struct host_info *);
static size_t http_read_msg(struct http_msg *, struct host_info *);

static size_t rebuild_content(struct http_msg *, msg_chunk *, size_t);
static char  *http_build_request(struct s_url *);
static void   http_build_hdr(struct http_msg *, char *, ...);
static void   http_build_default_hdr_list(struct s_url *, struct http_msg *);
static struct http_msg *http_make_req(struct s_url *);


static sigjmp_buf btf;


static void h_sigalarm(int no)
{

 siglongjmp(btf, 1);
}


char *http_strerror(int no)
{
 static char h_strerror[255];
 
 switch( no ) {
   case 0:
     strcpy(h_strerror, "0- no error");
     break;
   case 1:
     strcpy(h_strerror, "1- I/O error");
     break;
   case 2:
     strcpy(h_strerror, "2- unable to bind socket");
     break;
   case 3:
     strcpy(h_strerror, "3- unexpected EOF");
     break;
   case 4:
     strcpy(h_strerror, "4- bad message");
     break;
   case 5:
     strcpy(h_strerror, "5- invalid host");
     break;
   case 6:
     strcpy(h_strerror, "6- invalid url");
     break;
   case 7:
     strcpy(h_strerror, "7- connection refused");
     break;
   case 8:
     strcpy(h_strerror, "8- connection timeout");
     break;
   default:
     strcpy(h_strerror, "- undefined error");
     break;
   }

 return h_strerror;
} 


/*             *
 ** HOST_INFO **
 *             */


struct host_info *hostinfo_alloc()
{
 struct host_info *h;

 h = xmalloc(sizeof(struct host_info));
 h->h_fx = NULL;

 return h;
}


void hostinfo_free(struct host_info *h)
{
 
 free(h->hostname);
 free(h);
}


struct host_info *get_hostinfo(struct s_url *url)
{
 struct host_info *h;

 h = hostinfo_alloc();

 if( hostinfo_resolve_hostname(url->host, h) == -1 )
    return NULL;
 HOSTINFO_BUILD_PORT(h->port, url->port);

 return h;
}



static int hostinfo_resolve_hostname(char *hostname, struct host_info *m)
{
 struct hostent *hp; 

 hp = gethostbyname(hostname);
 if( hp == NULL ) {
    http_errno = HT_INVALIDHOST;
    return -1;
    }
 m->hostname  = strdup(hostname);
 inet_ntop(hp->h_addrtype, hp->h_addr, m->ip_addr_s, INET4_ADDRLEN);
 memcpy(&(m->ip_addr_i), hp->h_addr, sizeof(struct in_addr));

 return 0;
}


static FILE *hostinfo_connect(struct host_info *hi)
{
 int sk;
 struct sockaddr_in w;

 bzero(&w, sizeof(struct sockaddr_in));
 w.sin_family = AF_INET;
 w.sin_port   = htons(hi->port);
 w.sin_addr   = hi->ip_addr_i;

 sk = socket(PF_INET, SOCK_STREAM, 0);
 if( sk == -1 ) {
    http_errno = HT_SOCKET;
    return NULL;
    }
 while( 1 ) {
    if( sigsetjmp(btf, 1) == 0 ) {
       alarm(CONNECT_TIMEOUT);
       if( connect(sk, (struct sockaddr *)&w, sizeof(w)) == -1 ) {
          alarm(0);
          http_errno = HT_CONNREFUSED;
          return NULL;
          }
       alarm(0);
       break;
       }
    else {
       http_errno = HT_TIMEOUT;
       return NULL;
       }
    }

 hi->h_fx = fdopen(sk, "r+");

 return hi->h_fx;
}


static void hostinfo_close(struct host_info *h)
{
 
 if( h->h_fx != NULL )
    fclose(h->h_fx);
}


void hostinfo_print(struct host_info *h)
{

 printf("hostname: %s\n", h->hostname);
 printf("address : %s\n", h->ip_addr_s);
 printf("port    : %d\n", h->port);
}



/*             *
 ** HTTP SPEC **
 *             */


/*** MISC ***/


int http_init()
{
 struct sigaction anew, aold;

 anew.sa_handler = h_sigalarm;
 sigemptyset(&anew.sa_mask);
 anew.sa_flags = SA_RESTART;

 return (sigaction(SIGALRM, &anew, &aold));
}


int http_get_rcode(struct http_msg *msg)
{
 int code;
 char *p, *r;

 p = (msg->st_line)+8;
 while( *p == ' ' )
       p++;
 r = wstrndup(p, 3);

 code = atoi(r);
 free(r);

 return code;
}


void http_print_msg(struct http_msg *m)
{
 int i;
 
 printf("startline : %s\n", m->st_line);
 for(i=0; m->hdr_list[i] != NULL; i++)
     printf("header [%d]:  %s\n", i, m->hdr_list[i]);
}


void http_print_msg_content(struct http_msg *m)
{

 if( m->content_len > 0 )
    printf("content:\n%s\n", m->content);
 else
    printf("content:\n  (no content)\n");
}


/*** MEMORY ***/


struct http_msg *http_msg_alloc() 
{
 struct http_msg *tr;

 tr = xcalloc(1, sizeof(struct http_msg));
 tr->next_hdr = &(tr->hdr_list[0]);

 return tr;
} 


void http_free_msg(struct http_msg *m)
{
 int i;

 free(m->st_line);
 for(i=0; m->hdr_list[i] != NULL; i++)
     free(m->hdr_list[i]);
 if( m->content_len )
    free(m->content);
 free(m);
}
 

static void http_free_chunks(msg_chunk *m)
{
 msg_chunk *p, *q;

 p = m; 
 while( p != NULL ) {
       q = p->next;
       free(p->data);
       free(p);
       p = q;
       }
}


/*** I/O ***/


static size_t http_write_msg(struct http_msg *req, struct host_info *hi)
{
 int i;
 ssize_t sz, tsz;
 
 tsz = 0;

 sz = fwriten(req->st_line, strlen(req->st_line), 1, hi->h_fx);
#ifdef WDEBUG
     printf("request: %s\n", req->st_line);
#endif
 tsz += sz;
 if( sz != 1 )
    goto error;
 sz = fwriten("\r\n", 2, 1, hi->h_fx);
 tsz += sz;
 if( sz != 1 )
    goto error;
 for(i=0; req->hdr_list[i] != NULL; i++) {
     sz = fwriten(req->hdr_list[i], strlen(req->hdr_list[i]), 1, hi->h_fx);
#ifdef WDEBUG
     printf("request: %s\n", req->hdr_list[i]);
#endif
     tsz += sz;
     if( sz != 1 )
        goto error;
     sz = fwriten("\r\n", 2, 1, hi->h_fx);
     tsz += sz;
     if( sz != 1 )
       goto error;
     }
 sz = fwriten("\r\n", 2, 1, hi->h_fx);
 tsz += sz;
 if( sz != 1 )
    goto error;
 if( req->content_len ) {
    sz = fwriten(req->content, req->content_len, 1, hi->h_fx);
#ifdef WDEBUG
     printf("request: %s\n", req->content);
#endif
    tsz += sz;
    if( sz != 1 )
       goto error;
    }
 
 return tsz;

 error:
   if( sz == 0 ) {
      http_errno = HT_EOF;
      return -1;
      }
   else {
      http_errno = HT_IOERROR;
      return -1;
      }
}


static size_t http_read_msg(struct http_msg *rep, struct host_info *hi)
{
 int i;
 size_t sz, tsz;
 char *p;
 char  tmp[MAX_HDR_LINE_SIZE];
 msg_chunk *head_chnk;
 msg_chunk **c;

 head_chnk = NULL;

 rep->st_line = xmalloc(MAX_HDR_LINE_SIZE); 
 if( fgets(rep->st_line, MAX_HDR_LINE_SIZE, hi->h_fx) == NULL ) {
    if( feof(hi->h_fx) )
       http_errno = HT_EOF;
    else 
       http_errno = HT_IOERROR; 
    return -1;
    }
 p = strpbrk(rep->st_line, "\r\n");
 if( p == NULL ) {
    http_errno = HT_BADMSG;
    return -1;
    }
 *p = '\0';

 for(i=0; i<MAX_HDR_LIST_SIZE; i++) {
     if( fgets(tmp, MAX_HDR_LINE_SIZE, hi->h_fx) == NULL ) {
        if( feof(hi->h_fx) ) 
           return HT_EOF;
        else 
           return HT_IOERROR;
        return -1;
        }
     p = strpbrk(tmp, "\r\n");
     if( p == NULL ) {
        http_errno = HT_BADMSG;
        return -1;
        }
     if( p == tmp )
        break;
     *p = '\0';
     *(rep->next_hdr) = strdup(tmp);
     rep->next_hdr = rep->next_hdr++;
     } 

 tsz  = 0;
 c = &head_chnk;
 for(i=0; ;i++) {
     *c = xcalloc(1, sizeof(msg_chunk));
     (*c)->data = xmalloc(CHUNK_SIZE);
     sz = freadn((*c)->data, sizeof(char), CHUNK_SIZE-1, hi->h_fx);
     tsz += sz;
     if( sz == -1 ) {
        http_free_chunks(head_chnk);
        http_errno = HT_IOERROR;
        return -1;
        }
     (*c)->data[sz] = '\0'; 
     (*c)->chunk_size = sz;
     c = &(*c)->next;
     if( sz == 0 || sz != CHUNK_SIZE-1 ) 
        break;
     }

 rebuild_content(rep, head_chnk, tsz);
 http_free_chunks(head_chnk);

 return rep->content_len;
}


/*** BUILDER ***/


static size_t rebuild_content(struct http_msg *msg, msg_chunk *chk, size_t len)
{
 int i;
 char *d, *p;
 msg_chunk *c;

 if( len <= 0 ) 
    return len;

 c = chk;
 p = d = xmalloc(len+2);
 msg->content_len = len;

 for(i=0; c != NULL; i++, c=c->next) {
     memcpy(p, c->data, c->chunk_size);
     p += c->chunk_size;
     } 
/*
 * be flex friendly
 */
 *p++ = '\0';
 *p   = '\0';
 msg->content = d;

 return msg->content_len;
}


static char *http_build_request(struct s_url *url)
{
 char *m_req;
 char *query_line;

 if( url->method == M_GET ) {
    query_line = url_get_query_string(url);
    if( query_line == NULL )
       asprintf(&m_req, "GET %s %s", url->path, HTTP_VERSION);
    else { 
       asprintf(&m_req, "GET %s?%s %s", url->path, query_line, HTTP_VERSION);
       free(query_line);
       }
    }
 else
    asprintf(&m_req, "POST %s %s", url->path, HTTP_VERSION);

 return m_req;
}


static void http_build_hdr(struct http_msg *m, char *fmt, ...)
{
 va_list ap;
 
 va_start(ap, fmt);
 vasprintf(m->next_hdr, fmt, ap);
 va_end(ap);
 m->next_hdr = m->next_hdr++;
}


static void http_build_default_hdr_list(struct s_url *u, struct http_msg *m)
{
 char *cookie;

 http_build_hdr(m, "Host: %s", u->host);
 http_build_hdr(m, "User-Agent: %s", USER_AGENT);
 http_build_hdr(m, "Accept: */*");
 http_build_hdr(m, "Accept-Charset: ISO-8859-1");
 http_build_hdr(m, "Connection: close");
 if( cookie_tree != NULL ) {
    cookie = cookie_build(u->path, cookie_tree);
    if( cookie != NULL ) {
#ifdef WDEBUG
       printf("cookie: %s\n", cookie);
#endif
       http_build_hdr(m, "Cookie: %s", cookie);
       free(cookie);
       }
    }
 if( u->method == M_POST && m->content_len != 0) {
    http_build_hdr(m, "Content-Type: application/x-www-form-urlencoded");
    http_build_hdr(m, "Content-Length: %d", m->content_len );
    }
}

 
static struct http_msg *http_make_req(struct s_url *url)
{
 struct http_msg *m_req;

 m_req = http_msg_alloc();
 m_req->st_line = http_build_request(url);
 http_build_default_hdr_list(url, m_req);
 if( url->method == M_POST ) {
    m_req->content = url_get_query_string(url);
    if( m_req->content != NULL )
       m_req->content_len = strlen(m_req->content);
    }      

 return m_req;
}


char *http_get_hdr_value(struct http_msg *msg, char *hdr)
{
 int i, l;
 char *pt;
 
 l = strlen(hdr);
 for(i=0; msg->hdr_list[i] != NULL; i++) {
     if( strncasecmp(msg->hdr_list[i], hdr, l) == 0 ) {
        pt = msg->hdr_list[i]+l;
        SKIPSP(pt);
        if( *pt == ':' ) {
           pt++;
           SKIPSP(pt);
           }
        return strdup(pt);
        }
     }
 return NULL;
}


int http_cookie_insert(struct http_msg *msg, struct cook_tree *ct)
{
 int i;

 for(i=0; msg->hdr_list[i] != NULL; i++) {
     if( strncasecmp(msg->hdr_list[i], "set-cookie:", 11) == 0 )
        cookie_insert((msg->hdr_list[i])+11, ct);
     }
 return 0;
}


/*** ACTION ***/


struct http_msg *http_action(struct host_info *hi, struct s_url *url)
{
 int r;
 struct http_msg *reply, *request;

 request = http_make_req(url);
 if( request == NULL ) {
    http_errno = HT_INVALIDURL;
    return NULL;
    }

 if( hostinfo_connect(hi) == NULL ) {
    http_free_msg(request);  
    return NULL;
    }

 r = http_write_msg(request, hi);
 if( r <= 0 ) {
    http_free_msg(request);
    hostinfo_close(hi);
    return NULL;
    }
       
 reply = http_msg_alloc();
 r = http_read_msg(reply, hi);
 if( r < 0 ) {
    http_free_msg(request);  
    http_free_msg(reply);  
    hostinfo_close(hi);
    return NULL;
    }

 http_free_msg(request);
 hostinfo_close(hi);

 return reply;
}
