#ifndef _HTTP_H
#define _HTTP_H

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>

#include "url.h"
#include "cookies.h"

#define INET4_ADDRLEN      16

#define P_VERSION    "test 0.0.1"
#define USER_AGENT   "Mozilla/5.0 (X11; U; FreeBSD i386; en-US; rv=0.9.9) Gecko/20020312"

#define MAX_HDR_LIST_SIZE  64
#define MAX_HDR_LINE_SIZE  512

#define HTTP_VERSION      "HTTP/1.0"

#define HT_IOERROR          1
#define HT_SOCKET           2
#define HT_EOF              3
#define HT_BADMSG           4
#define HT_INVALIDHOST      5
#define HT_INVALIDURL       6 /* check url_errno */
#define HT_CONNREFUSED      7
#define HT_TIMEOUT          8

#define FORBIDDEN_LINK     -1 

#define M_GET               0
#define M_POST              1

struct host_info {
   FILE          *h_fx;
   struct in_addr ip_addr_i;
   char           ip_addr_s[INET4_ADDRLEN];
   char          *hostname;
   unsigned short port;
   };

struct http_msg_chunk {
   char                  *data;
   size_t                 chunk_size;
   struct http_msg_chunk *next;
   };
typedef struct http_msg_chunk msg_chunk;

struct http_msg {
   int           method;
   char         *st_line;
   char         *hdr_list[MAX_HDR_LIST_SIZE];
   char        **next_hdr;
   char         *content;
   size_t        content_len;
   };


struct host_info *hostinfo_alloc(void);
void   hostinfo_free(struct host_info *);
struct host_info *get_hostinfo(struct s_url *);
void   hostinfo_print(struct host_info *h);

int              http_init(void);
void             http_print_msg(struct http_msg *);
void             http_print_msg_content(struct http_msg *);
struct http_msg *http_msg_alloc(void);
void             http_free_msg(struct http_msg *);

void http_set_method(int);
int  http_get_method(void);
char *http_get_hdr_value(struct http_msg *, char *);
int  http_cookie_insert(struct http_msg *, struct cook_tree *);
int  http_get_rcode(struct http_msg *);

struct http_msg *http_action(struct host_info *, struct s_url *);

char *http_strerror(int);

int http_errno;

#endif
