/*
 * rfc 1630
 * rfc 1738
 * rfc 1808
 * rfc 2396
 * rfc 2616
 *
 */
#ifndef _URL_H
#define _URL_H

#define UE_NOSTRING         1
#define UE_TRAILINGPERCENT  2
#define UE_UNSUPPORTED      3
#define UE_NOHOST           4
#define UE_BADLP            5
#define UE_BADPORT          6
#define UE_QUERYLEN         7
#define UE_FORBIDDEN        8
#define UE_UNDEFINED        255

#define MAX_QUERYPAIR    128

#define ABSOLUTE_URL     0
#define NETWORK_URL      1
#define RELATIVE_URL     2
#define RECALIBRATED_URL 3

#define USER_CONTEXT     0
#define HREF_CONTEXT     1

#define PROTO_HTTP          1
#define PROTO_HTTPS         2
#define PROTO_UNSUPPORTED  -1

#define HTTP_DEFAULT_PORT   80

struct s_param {
   char *name;
   char *value;
   };

struct s_query {
   int    sz;
   struct s_param *pair[MAX_QUERYPAIR];
   };
  
struct s_url {
   int    method;
   int    type;
   int    scheme;   /* 0 = not defined */
   char  *user;
   char  *pass;
   char  *host;
   int    port;
   char  *path;
   struct s_query *query;
   char  *frag;
   };

struct e_url {
   struct s_url *url;
   struct e_url *next;
   };

struct url_list {
   struct e_url    *head;
   struct e_url    *tail;
   int sz;
   };

struct url_list *url_list_alloc(void);
void url_list_free(struct url_list *);

struct e_url *e_url_alloc(void);
void e_url_free(struct e_url *);

int  url_list_stradd(struct url_list *, char *);
void url_list_urladd(struct url_list *, struct s_url *);

void url_list_print(struct url_list *);

struct s_url *url_alloc(void);
void          url_free(struct s_url *);

void url_print(struct s_url *);

struct s_url *url_build(char *purl);
char         *url_tostr(struct s_url *);
int           url_recalibrate(struct s_url *, struct s_url *);
char         *url_get_query_string(struct s_url *);

void url_querypair_add(char *, char *, int, struct s_url *);

int url_recalibrate(struct s_url *, struct s_url *);

char *url_strerror(int);

int url_set_context(int);
int url_get_context(void);
struct s_url *url_dup(struct s_url *);
struct s_url *url_redirect(struct s_url *, char *);

int url_errno;

#endif

